// Comtrol RocketPort Infinity/Express custom version of setserial.
//
// This program does basically the same things as setserial except:
//  * it ignores fields that aren't used by the Inf/Exp boards
//  * it supports setting the transceiver selection field (RS232/RS422/RS485/...)
//  * it will optionally create the appropriate /dev/RPnn devices
//  * it can print out board model/version information

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <unistd.h>
#include <linux/serial.h>

#define MKDEV(a,b) ((int)((((a) & 0xff) << 8) | ((b) & 0xff)))

// custom ioctl commands
#define RCKP_GET_PORTS          0x00525004
#define RCKP_GET_VERSION        0x00525006

// these flag definitions have to be kept in sync with those in
// rocket2.h Note: rp infinity driver supports an "extra"
// non-standard ASYNC_SPD_WARP13 flag value

#define ASYNC_SPD_WARP13       (ASYNC_SPD_HI|ASYNC_SPD_VHI|ASYNC_SPD_SHI)

#if !defined(ASYNCB_LAST_USER)
# define ASYNCB_LAST_USER 15
#endif

#define ROCKETB_RTS_TOGGLE      (ASYNCB_LAST_USER+1)
#define ROCKETB_IS_SMPTE        (ASYNCB_LAST_USER+2)
#define ROCKETB_MODE            (ASYNCB_LAST_USER+3)
#define ROCKETB_MODE_WIDTH      3
#define ROCKETB_LAST            (ROCKETB_MODE+ROCKETB_MODE_WITDH-1)

#define ROCKET_RTS_TOGGLE     (1U << ROCKETB_RTS_TOGGLE)
#define ROCKET_IS_SMPTE       (1U << ROCKETB_IS_SMPTE)

#define ROCKET_MODE_MASK      ((~((~0U)<<ROCKETB_MODE_WIDTH)) << ROCKETB_MODE)
#define ROCKET_MODE_RS232     (0 << ROCKETB_MODE)
#define ROCKET_MODE_RS422     (1 << ROCKETB_MODE)
#define ROCKET_MODE_RS485     (2 << ROCKETB_MODE) // 485 2-wire
#define ROCKET_MODE_RS485FDS  (3 << ROCKETB_MODE) // 485 4-wire slave
#define ROCKET_MODE_RS485FDM  (4 << ROCKETB_MODE) // 485 4-wire master

typedef struct {
    unsigned long model;
    char modelString[80];
    unsigned long numPorts;
    int startingPortNumber;
} rocketModel_t;

struct rocket_ports {
    int tty_major;
    int callout_major;
    rocketModel_t rocketModel[8];
};

struct rocket_version {
    char rocket_version[32];
    char rocket_date[32];
    char reserved[64];
};

//-----------------------------------------------------------------------------------

#define VERSION_STR "1.08"

char *progname;

int     verbosity = 1;          /* 1 = normal, 0=boot-time, 2=everything */
int     verbose_flag = 0;       /* print results after setting a port */
int     quiet_flag = 0;
int     old_device_name = 0;

#define CMD_FLAG        1
#define CMD_DELAY       2
#define CMD_WAIT        3
#define CMD_BBASE       4
#define CMD_DIVSR       5


#define FLAG_CAN_INVERT 0x0001
#define FLAG_NEED_ARG   0x0002
#define FLAG_READ_ONLY  0x0004

struct flag_type_table {
    int cmd;
    char        *name;
    int bits;
    int mask;
    int level;
    int flags;
} flag_type_tbl[] = {
    {CMD_FLAG,   "rtsToggle",            ROCKET_RTS_TOGGLE,     ROCKET_RTS_TOGGLE,     0,    FLAG_CAN_INVERT},
    {CMD_FLAG,   "spd_normal",           0,                     ASYNC_SPD_MASK,        0,    0},
    {CMD_FLAG,   "spd_hi",               ASYNC_SPD_HI,          ASYNC_SPD_MASK,        0,    0},
    {CMD_FLAG,   "spd_vhi",              ASYNC_SPD_VHI,         ASYNC_SPD_MASK,        0,    0},
    {CMD_FLAG,   "spd_shi",              ASYNC_SPD_SHI,         ASYNC_SPD_MASK,        0,    0},
    {CMD_FLAG,   "spd_warp",             ASYNC_SPD_WARP,        ASYNC_SPD_MASK,        0,    0},
    {CMD_FLAG,   "spd_warp13",           ASYNC_SPD_WARP13,      ASYNC_SPD_MASK,        0,    0},
    {CMD_FLAG,   "spd_cust",             ASYNC_SPD_CUST,        ASYNC_SPD_MASK,        0,    0},
    {CMD_FLAG,   "rs232",                ROCKET_MODE_RS232,     ROCKET_MODE_MASK,      0,    0},
    {CMD_FLAG,   "rs422",                ROCKET_MODE_RS422,     ROCKET_MODE_MASK,      0,    0},
    {CMD_FLAG,   "rs485_2-wire",         ROCKET_MODE_RS485,     ROCKET_MODE_MASK,      0,    0},
    {CMD_FLAG,   "rs485_4-wire_slave",   ROCKET_MODE_RS485FDS,  ROCKET_MODE_MASK,      0,    0},
    {CMD_FLAG,   "rs485_4-wire_master",  ROCKET_MODE_RS485FDM,  ROCKET_MODE_MASK,      0,    0},
    {CMD_FLAG,   "SMPTE",                ROCKET_IS_SMPTE,       ROCKET_IS_SMPTE,       0,    FLAG_READ_ONLY},
    {CMD_FLAG,   "hup_notify",           ASYNC_HUP_NOTIFY,      ASYNC_HUP_NOTIFY,      0,    FLAG_CAN_INVERT},
    {CMD_DELAY,  "close_delay",          0,                     0,                     0,    FLAG_NEED_ARG},
    {CMD_WAIT,   "closing_wait",         0,                     0,                     0,    FLAG_NEED_ARG},
    {CMD_BBASE,  "baud_base",            0,                     0,                     0,    FLAG_NEED_ARG},
    {CMD_DIVSR,  "divisor",              0,                     0,                     0,    FLAG_NEED_ARG},
    {0,          0,                      0,                     0,                     0,    0},
};


int atonum(char *s)
{
    int n;

    while (*s == ' ')
        s++;
    if (strncmp(s, "0x", 2) == 0 || strncmp(s, "0X", 2) == 0)
        sscanf(s + 2, "%x", &n);
    else if (s[0] == '0' && s[1])
        sscanf(s + 1, "%o", &n);
    else
        sscanf(s, "%d", &n);
    return n;
}

void print_flags(struct serial_struct *serinfo,
                 char *prefix, char *postfix)
{
    struct      flag_type_table *p;
    int flags;
    int first = 1;

    flags = serinfo->flags;

    for (p = flag_type_tbl; p->name; p++) {
        if (p->cmd != CMD_FLAG)
            continue;
        if (verbosity < p->level)
            continue;
        if ((flags & p->mask) == p->bits) {
            if (first) {
                printf("%s", prefix);
                first = 0;
            } else
                printf(" ");
            printf("%s", p->name);
        }
    }

    if (!first)
        printf("%s", postfix);
}

void get_rocket(char *device)
{
    struct serial_struct config;
    int fd;
    char        buf1[40];

    if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
        perror(device);
        return;
    }
    if (ioctl(fd, TIOCGSERIAL, &config) < 0) {
        perror("Cannot get serial info");
        close(fd);
        return;
    }
    printf("flags=%08x\n", config.flags);
    if (verbosity==2) {
        printf("%s, Line %d, Port %d, Iomem %p\n", device, config.line, config.port, config.iomem_base);
        if (config.closing_wait == ASYNC_CLOSING_WAIT_INF)
            strcpy(buf1, "infinte");
        else if (config.closing_wait == ASYNC_CLOSING_WAIT_NONE)
            strcpy(buf1, "none");
        else
            sprintf(buf1, "%d", config.closing_wait);
        printf("\tclose_delay: %d, closing_wait: %s\n", config.close_delay, buf1);
        printf("\tbaud_base: %d, divisor: %d\n", config.baud_base, config.custom_divisor);
        print_flags(&config, "\tFlags: ", "");
        printf("\n\n");
    } else {
        printf("%s, ", device);
        print_flags(&config, "Flags: ", "");
        printf("\n");
    }
    close(fd);
}


void set_rocket(char *device, char ** arg)
{
    struct serial_struct  old_config, new_config;
    struct flag_type_table *p;
    int fd;
    int do_invert = 0;
    char        *word;

    if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
        if (verbosity==0 && errno==ENOENT)
            exit(201);
        perror(device);
        exit(201);
    }
    if (fcntl(fd, F_SETFL, 0) < 0) {
        perror("fcntl");
        exit(1);
    }
    if (ioctl(fd, TIOCGSERIAL, &old_config) < 0) {
        perror("Cannot get rocket info");
    }
    new_config = old_config;
    //printf("old flags=%08x\n", new_config.flags);
    while (*arg) {
        do_invert = 0;
        word = *arg++;
        if (*word == '^') {
            do_invert++;
            word++;
        }
        for (p = flag_type_tbl; p->name; p++) {
            if (!strcasecmp(p->name, word))
                break;
        }
        if (!p->name) {
            fprintf(stderr, "Invalid flag: %s\n", word);
            exit(1);
        }
        if (p->flags & FLAG_READ_ONLY) {
            fprintf(stderr, "This flag is read only: %s\n", word);
            exit(1);
        }
        if (do_invert && !(p->flags & FLAG_CAN_INVERT)) {
            fprintf(stderr, "This flag can not be inverted: %s\n", word);
            exit(1);
        }
        if ((p->flags & FLAG_NEED_ARG) && !*arg) {
            fprintf(stderr, "Missing argument for %s\n", word);
            exit(1);
        }
        if ((p->mask == ROCKET_MODE_MASK) && (new_config.flags & ROCKET_IS_SMPTE)) {
            fprintf(stderr, "Port interface hardware of SMPTE cards can not be changed: %s\n", word);
            exit(1);
        }
        switch (p->cmd) {
        case CMD_FLAG:
            new_config.flags &= ~p->mask;
            if (!do_invert)
                new_config.flags |= p->bits;
            break;
        case CMD_DELAY:
            new_config.close_delay = atonum(*arg++);
            break;
        case CMD_BBASE:
            new_config.baud_base = atonum(*arg++);
            break;
        case CMD_DIVSR:
            new_config.custom_divisor = atonum(*arg++);
            break;
        case CMD_WAIT:
            if (!strcasecmp(*arg, "infinite"))
                new_config.closing_wait = ASYNC_CLOSING_WAIT_INF;
            else if (!strcasecmp(*arg, "none"))
                new_config.closing_wait = ASYNC_CLOSING_WAIT_NONE;
            else
                new_config.closing_wait = atonum(*arg);
            arg++;
            break;
        default:
            fprintf(stderr, "Internal error: unhandled cmd #%d\n", p->cmd);
            exit(1);
        }
    }
    //printf("new flags=%08x\n", new_config.flags);
    if (ioctl(fd, TIOCSSERIAL, &new_config) < 0) {
        perror("Cannot set rocket info");
        exit(1);
    }
    close(fd);
    if (verbose_flag)
        get_rocket(device);
}

void make_device(char *pathname, int mode, int major, int minor,
                 gid_t tty_group)
{
    struct stat statbuf;
    dev_t       device;

    device = MKDEV(major, minor);
    if (lstat(pathname, &statbuf) == 0) {
        if (device == statbuf.st_rdev)
            return;
        unlink(pathname);
    }
    if (mknod(pathname, mode, device) < 0) {
        perror(pathname);
    }
    chown(pathname, 0, tty_group);
}

void mkdev_rocket(char *device)
{
    struct rocket_version rocketVer;
    struct rocket_ports rocketInfo;
    int fd, board, line;
    gid_t tty_group = 0;
    struct group *grp;
    char pathname[64];
    mode_t mode;

    grp = getgrnam("tty");
    if (grp)
        tty_group = grp->gr_gid;

    if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
        perror(device);
        return;
    }
    if (ioctl(fd, RCKP_GET_VERSION, &rocketVer) < 0) {
        perror("Cannot get rocket driver version");
        return;
    }
    if (ioctl(fd, RCKP_GET_PORTS, &rocketInfo) < 0) {
        perror("Cannot get rocket ports structure");
        close(fd);
        return;
    }

    close(fd);
    mode = S_IFCHR | 0620;
    umask(0);
    fflush(stdout);

    printf("RocketPort Infinity device driver module, version %s, %s\n\n",
           rocketVer.rocket_version, rocketVer.rocket_date);

    // Walk through the model info struct a board at a time, creating
    // the /dev/ttyRPx devices

    for (board = 0; board < 4; board++) {
        if (!rocketInfo.rocketModel[board].model)
            break;

        printf("Installing %s, creating /dev/ttyRP%d - %lu\n",
               rocketInfo.rocketModel[board].modelString,
               rocketInfo.rocketModel[board].startingPortNumber,
               rocketInfo.rocketModel[board].startingPortNumber + rocketInfo.rocketModel[board].numPorts - 1);

        for (line = rocketInfo.rocketModel[board].startingPortNumber;
                line < rocketInfo.rocketModel[board].startingPortNumber + rocketInfo.rocketModel[board].numPorts;
                line++) {
            //  Create the /dev/ttyRP0 device
            sprintf(pathname, "/dev/ttyRP%d", line);
            //printf("DEBUG: Creating %s \n", pathname);
            make_device(pathname, mode, rocketInfo.tty_major, line, tty_group);
        }
    }
}

void usage()
{
    fprintf(stderr, "setinfinity Version %s\n\n", VERSION_STR);
    fprintf(stderr,
            "usage: %s rocket-device [cmd1 [arg]] ... \n\n", progname);
    fprintf(stderr, "Available commands: (* = Takes an argument)\n");
    fprintf(stderr, "\t\t(^ = can be preceded by a '^' to turn off the option)\n");
    fprintf(stderr, "\t* close_delay\tset the amount of time (in 1/100 of a\n");
    fprintf(stderr, "\t\t\t\tsecond) that DTR should be kept low\n");
    fprintf(stderr, "\t\t\t\twhile being closed\n");
    fprintf(stderr, "\t* closing_wait\tset the amount of time (in 1/100 of a\n");
    fprintf(stderr, "\t\t\t\tsecond) that the rocket port should wait for\n");
    fprintf(stderr, "\t\t\t\tdata to be drained while being closed\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t^ rtsToggle\tEnable the RTS toggle flag on the port\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t  rs232\t\tSet port interface hardware to RS-232\n");
    fprintf(stderr, "\t  rs422\t\tSet port interface hardware to RS-422\n");
    fprintf(stderr, "\t  rs485_2-wire\t\tSet port interface hardware to RS-485 2-wire\n");
    fprintf(stderr, "\t  rs485_4-wire_slave\t\tSet port interface hardware to RS-485 4-wire Slave\n");
    fprintf(stderr, "\t  rs485_4-wire_master\t\tSet port interface hardware to RS-485 4-wire Master\n");
    fprintf(stderr, "\n");
#if 0
    fprintf(stderr, "\t^ sak\t\tset the break key as the Secure Attention Key\n");
#endif
    fprintf(stderr, "\t  spd_hi\tuse 56kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_vhi\tuse 115kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_shi\tuse 230kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_warp\tuse 460kb instead of 38.4kb\n");
    fprintf(stderr, "\t  spd_warp13\tuse 921kb instead of 38.4kb\n");//BKA
    fprintf(stderr, "\t  spd_normal\tuse 38.4kb when a baud rate of 38.4kb is selected\n");
    fprintf(stderr, "\n");
    exit(1);
}

int main(int argc, char **argv)
{
    int get_flag = 0;
    int mkdev_flag = 0;
    int c;
    extern int optind;
    extern char *optarg;

    progname = argv[0];
    if (argc == 1)
        usage();
    while ((c = getopt(argc, argv, "abgoqvMVW")) != EOF) {
        switch (c) {
        case 'a':
            verbosity = 2;
            break;
        case 'b':
            verbosity = 0;
            break;
        case 'q':
            quiet_flag++;
            break;
        case 'v':
            verbose_flag++;
            break;
        case 'g':
            get_flag++;
            mkdev_flag = 0;
            break;
        case 'o':
            old_device_name++;
            break;
        case 'M':
            mkdev_flag++;
            get_flag = 0;
            break;
        case 'V':
            fprintf(stderr, "setrocket version %s\n", VERSION_STR);
            exit(0);
        default:
            usage();
        }
    }
    if (get_flag) {
        argv += optind;
        while (*argv)
            get_rocket(*argv++);
        exit(0);
    }
    if (argc == optind)
        usage();
    if (mkdev_flag) {
        mkdev_rocket(argv[optind]);
        exit(0);
    }
    if (argc-optind == 1)
        get_rocket(argv[optind]);
    else
        set_rocket(argv[optind], argv+optind+1);
    exit(0);
}

